# -*- coding: utf-8 -*-
"""
"""
from pywapor.collect.product import LANDSAT, CHIRPS, ERA5, GEOS5, GLOBCOVER, MERRA2, MODIS, TERRA, SRTM, STATICS, SENTINEL2, SENTINEL3, VIIRSL1, COPERNICUS

__all__ = ["LANDSAT", "CHIRPS", "ERA5", "GEOS5", "GLOBCOVER", "MERRA2", "MODIS", "TERRA", "SRTM", "STATICS", "SENTINEL2", "SENTINEL3", "VIIRSL1", "COPERNICUS"]

__version__ = '0.1'
